package de.storchp.fdroidbuildstatus.monitor

import android.app.job.JobInfo
import android.app.job.JobParameters
import android.app.job.JobScheduler
import android.app.job.JobService
import android.content.ComponentName
import android.content.Context
import android.util.Log
import androidx.core.content.ContextCompat
import androidx.work.Configuration
import dagger.hilt.android.AndroidEntryPoint
import de.storchp.fdroidbuildstatus.R
import de.storchp.fdroidbuildstatus.adapter.ApiResponse
import de.storchp.fdroidbuildstatus.adapter.db.DbAdapter
import de.storchp.fdroidbuildstatus.adapter.fdroid.BuildResult
import de.storchp.fdroidbuildstatus.adapter.fdroid.FdroidClient
import de.storchp.fdroidbuildstatus.adapter.fdroid.RunningResult
import de.storchp.fdroidbuildstatus.model.App
import de.storchp.fdroidbuildstatus.model.AppNotification
import de.storchp.fdroidbuildstatus.model.BuildCycle
import de.storchp.fdroidbuildstatus.utils.NotificationUtils
import de.storchp.fdroidbuildstatus.utils.PreferencesService
import javax.inject.Inject

@AndroidEntryPoint
class MonitorJobService : JobService() {

    @Inject
    lateinit var preferencesService: PreferencesService

    @Inject
    lateinit var dbAdapter: DbAdapter

    @Inject
    lateinit var fdroidClient: FdroidClient

    init {
        val builder: Configuration.Builder = Configuration.Builder()
        builder.setJobSchedulerJobIdRange(0, 1000)
    }

    override fun onStartJob(params: JobParameters): Boolean {
        Log.i(TAG, "MonitorService job started")
        val buildRunType = preferencesService.getBuildCycleCheck()
        if (buildRunType === BuildCycle.BUILD) {
            fdroidClient.getBuild { response: ApiResponse<BuildResult?> ->
                Log.d(TAG, "build response: " + response.isSuccess)
                if (response.isSuccess) {
                    processBuildRun(response.value(), dbAdapter, buildRunType)
                } else {
                    Log.e(TAG, "MonitorService failed: " + response.errorMessage())
                }
                jobFinished(params, false)
            }
        } else if (buildRunType === BuildCycle.RUNNING) {
            fdroidClient.getRunning { response: ApiResponse<RunningResult?> ->
                Log.d(TAG, "running response: " + response.isSuccess)
                if (response.isSuccess) {
                    val result = response.value()
                    if (result is BuildResult) {
                        processBuildRun(result, dbAdapter, buildRunType)
                    }
                } else {
                    Log.e(TAG, "MonitorService failed: " + response.errorMessage())
                }
                jobFinished(params, false)
            }
        }
        return true
    }

    private fun processBuildRun(
        buildRun: BuildResult?,
        dbAdapter: DbAdapter?,
        buildCycle: BuildCycle
    ) {
        val oldBuildRun = dbAdapter!!.loadBuildRuns()[buildCycle]
        Log.d(
            TAG,
            "oldBuild Last-Modified: " + (oldBuildRun?.lastModified
                ?: "") + ", newBuildRun Last-Modified: " + buildRun!!.lastModified
        )
        if (oldBuildRun == null || oldBuildRun.lastModified != buildRun.lastModified) {
            dbAdapter.saveBuildRun(buildRun, buildCycle)
            val favourites = dbAdapter.favourites
            if (preferencesService.isNotifyFavouritesOnly()) {
                notifyFavourites(buildRun, dbAdapter, buildCycle, favourites)
            } else {
                NotificationUtils.createNewBuildNotification(
                    this@MonitorJobService,
                    getString(R.string.new_build_notification)
                )
            }
        }
    }

    private fun notifyFavourites(
        buildRun: BuildResult?,
        dbAdapter: DbAdapter?,
        buildCycle: BuildCycle,
        favourites: Map<String, App>
    ) {
        val notifiedApps = dbAdapter!!.getNotificationsFor(buildCycle, buildRun!!.startTimestamp)
        val newFavoriteApps = buildRun.allBuilds()
            .map {
                AppNotification(
                    it.id,
                    it.versionCode
                )
            }
            .filter { a: AppNotification? ->
                favourites.containsKey(
                    a!!.id
                ) && !notifiedApps.contains(a)
            }
            .toSet()
        if (newFavoriteApps.isNotEmpty()) {
            if (newFavoriteApps.size <= 3) {
                val names = newFavoriteApps
                    .mapNotNull { a: AppNotification? -> favourites[a!!.id] }
                    .map(App::displayName)
                    .joinToString(", ")
                NotificationUtils.createNewBuildNotification(
                    this@MonitorJobService,
                    getString(R.string.new_build_notification_fav_names, names)
                )
            } else {
                NotificationUtils.createNewBuildNotification(
                    this@MonitorJobService,
                    getString(R.string.new_build_notification_favs, newFavoriteApps.size)
                )
            }
            notifiedApps.addAll(newFavoriteApps)
            dbAdapter.saveNotifications(buildCycle, buildRun.startTimestamp, notifiedApps)
        }
    }

    override fun onStopJob(params: JobParameters): Boolean {
        return true
    }

    companion object {
        private val TAG = MonitorJobService::class.java.simpleName
        private const val JOB_ID = 1000
        fun schedule(context: Context, preferencesService: PreferencesService) {
            val jobScheduler = ContextCompat.getSystemService(context, JobScheduler::class.java)!!
            jobScheduler.cancel(JOB_ID) // cancel previous scheduled job
            if (!preferencesService.isUpdateCheckEnabled()) {
                Log.i(TAG, "MonitorService disabled")
                return
            }
            if (!NotificationUtils.areNotificationsEnabled(context)) {
                Log.i(TAG, "Notifications disabled, disable MonitorService")
                preferencesService.setUpdateCheckEnabled(false)
                return
            }
            val builder =
                JobInfo.Builder(JOB_ID, ComponentName(context, MonitorJobService::class.java))
            val updateInterval = preferencesService.getUpdateInterval()
            builder.setPeriodic(updateInterval)
                .setRequiredNetworkType(JobInfo.NETWORK_TYPE_ANY)
            builder.setRequiresBatteryNotLow(true)
            jobScheduler.schedule(builder.build())
            Log.i(TAG, "MonitorService scheduled for $updateInterval millis interval")
        }
    }
}